@@ -1,13 +1,16 @@
 import os
+import time
 from getpass import getpass
 
 # from cloudmesh.common.debug import VERBOSE
 from cloudmesh.burn.Imager import Imager
 from cloudmesh.burn.burner.Burner import Burner
+from cloudmesh.burn.burner.RaspberryBurner import Burner as RaspberryBurner
 from cloudmesh.burn.burner.raspberryos import MultiBurner
 from cloudmesh.burn.image import Image
 from cloudmesh.burn.network import Network
 from cloudmesh.burn.sdcard import SDCard
+from cloudmesh.burn.ubuntu.configure import Configure
 from cloudmesh.burn.util import os_is_linux
 from cloudmesh.burn.util import os_is_mac
 from cloudmesh.burn.util import os_is_pi
@@ -17,11 +20,17 @@ from cloudmesh.common.parameter import Parameter
 from cloudmesh.common.security import generate_strong_pass
 from cloudmesh.common.util import Console
 from cloudmesh.common.util import yn_choice
+from cloudmesh.common.util import banner
+from cloudmesh.inventory.inventory import Inventory
 from cloudmesh.shell.command import PluginCommand
 from cloudmesh.shell.command import command
 from cloudmesh.shell.command import map_parameters
 from cloudmesh.burn.usb import USB
 from cloudmesh.common.debug import VERBOSE
+from cloudmesh.burn.wifi.ssid import get_ssid
+from cloudmesh.common.Host import Host
+from cloudmesh.common.util import path_expand
+from cloudmesh.common.Shell import Shell
 
 
 class BurnCommand(PluginCommand):
@@ -39,6 +48,19 @@ class BurnCommand(PluginCommand):
                        [--wifipassword=PSK]
                        [--bs=BLOCKSIZE]
                        [--dryrun]
+                       [--no_diagram]
+              burn ubuntu NAMES [--inventory=INVENTORY] [--ssid=SSID]
+              [--wifipassword=PSK] [-v] --device=DEVICE [--country=COUNTRY]
+              [--upgrade]
+              burn raspberry NAMES --device=DEVICE
+                                  [--inventory=INVENTORY]
+                                  [--ssid=SSID]
+                                  [--wifipassword=PSK]
+                                  [--country=COUNTRY]
+                                  [--password=PASSWORD]
+                                  [-v]
+                                  [-f]
+              [--timezone=TIMEZONE]
               burn firmware check
               burn firmware update
               burn install
@@ -87,6 +109,7 @@ class BurnCommand(PluginCommand):
                        [--ip=IP]
                        [--key=KEY]
                        [--keyboard=COUNTRY]
+                       [--cmdline=CMDLINE]
               burn enable ssh
               burn wifi --ssid=SSID [--passwd=PASSWD] [--country=COUNTRY]
               burn check [--device=DEVICE]
@@ -276,6 +299,7 @@ class BurnCommand(PluginCommand):
                              [--key=KEY]
                              [--mount=MOUNTPOINT]
                              [--keyboard=COUNTRY]
+                             [--cmdline=CMDLINE]
 
                     Sets specific values on the sdcard after it
                     has ben created with the create, copy or sdcard
@@ -344,7 +368,10 @@ class BurnCommand(PluginCommand):
                        "inventory",
                        "name",
                        "bs",
-                       "set_passwd")
+                       "set_passwd",
+                       "cmdline",
+                       "upgrade",
+                       "no_diagram")
 
         # arguments.MOUNTPOINT = arguments["--mount"]
         arguments.FORMAT = arguments["--format"]
@@ -389,12 +416,166 @@ class BurnCommand(PluginCommand):
 
             from cloudmesh.burn.gui import Gui
             VERBOSE(arguments)
-            g = Gui(hostname=arguments.hostname, ip=arguments.ip, dryrun=arguments.dryrun)
+            g = Gui(hostname=arguments.hostname,
+                    ip=arguments.ip,
+                    dryrun=arguments.dryrun,
+                    no_diagram=arguments.no_diagram)
 
             g.run()
 
             return ""
 
+        elif arguments.raspberry:
+            banner(txt="RaspberryOS Burn", figlet=True)
+
+            names = Parameter.expand(arguments.NAMES)
+            manager, workers = Host.get_hostnames(names)
+            ssid = arguments['--ssid']
+            wifipasswd = arguments['--wifipassword']
+
+            if arguments.inventory:
+                burner = RaspberryBurner(inventory=arguments.inventory)
+            else:
+                if workers:
+                    worker_base_name = ''.join(
+                        [i for i in workers[0] if not i.isdigit()])
+
+                cluster_name = manager or worker_base_name
+                inventory = path_expand(f'~/.cloudmesh/inventory-{cluster_name}.yml')
+
+                if not os.path.exists(inventory) or arguments['-f']:
+                    if not manager:
+                        Console.error("No inventory found. Can not create an "
+                                      "inventory without a "
+                                      "manager.")
+                        return ""
+
+                    _build_default_inventory(filename=inventory,
+                                             manager=manager, workers=workers)
+
+                burner = RaspberryBurner(inventory=inventory)
+
+                if manager:
+                    if not ssid:
+                        ssid = get_ssid()
+                        if ssid == "":
+                            Console.info('Could not determine SSID, skipping wifi '
+                                         'config')
+                    if not wifipasswd and not ssid == "":
+                        wifipasswd = getpass(f"Using --SSID={ssid}, please "
+                                             f"enter wifi password:")
+
+            execute("burn raspberry", burner.multi_burn(
+                names=arguments.NAMES,
+                devices=arguments.device,
+                verbose=arguments['-v'],
+                password=arguments['--password'],
+                ssid=ssid,
+                wifipasswd=wifipasswd,
+                country=arguments['--country']
+            ))
+            return ""
+
+        elif arguments.ubuntu:
+            banner(txt="Ubuntu Burn with cloud-init", figlet=True)
+            names = Parameter.expand(arguments.NAMES)
+            if len(Parameter.expand(arguments.device)) > 1:
+                Console.error("Too many devices specified. Please only specify one")
+                return ""
+
+            if arguments.inventory:
+                c = Configure(inventory=arguments.inventory, debug=arguments['-v'])
+                inv = Inventory(filename=arguments.inventory)
+            else:
+                c = Configure(debug=arguments['-v'])
+                inv = Inventory()
+
+            # Probably not the best way to get the tag, but we assume all tags
+            # are the same for each row for now
+            if inv.has_host(names[0]):
+                tag = inv.get(name=names[0], attribute='tag')
+            else:
+                Console.error(f'Could not find {names[0]} in inventory {inv.filename}')
+                return ""
+            if 'ubuntu' not in tag:
+                Console.error("This command only supports burning ubuntu cards")
+                return ""
+            sdcard = SDCard(card_os="ubuntu")
+
+            # Code below taken from arguments.sdcard
+            try:
+                USB.check_for_readers()
+            except Exception as e:
+                print()
+                Console.error(e)
+                print()
+                return ""
+
+            # determine if we are burning a manager, as this needs to be done
+            # first to get the ssh public key
+            manager = False
+            for name in names:
+                if not inv.has_host(name):
+                    Console.error(f'Could not find {name} in inventory {inv.filename}')
+                    return ""
+                service = inv.get(name=name, attribute='service')
+                if service == 'manager' and not manager:
+                    manager = name
+                    # make manager first in names
+                    names.remove(name)
+                    names.insert(0, name)
+                elif service == 'manager' and manager:
+                    raise Exception('More than one manager detected in NAMES')
+
+            for name in names:
+                if not yn_choice(f'Is the card to be burned for {name} inserted?'):
+                    if not yn_choice(f"Please insert the card to be burned for {name}. "
+                                     "Type 'y' when done or 'n' to terminante"):
+                        Console.error("Terminating: User Break")
+                        return ""
+
+                service = inv.get(name=name, attribute='service')
+                # Make sure bridge is only enabled if WiFi enabled
+                if service == 'manager':
+                    services = inv.get(name=name, attribute='services')
+                    if 'bridge' in services and not arguments.ssid:
+                        Console.error('Service bridge can only be configured if WiFi'
+                                      ' is enabled with --ssid and --wifipassword')
+                        return ""
+                    else:
+                        enable_bridge = 'bridge' in services
+
+                Console.info(f'Burning {name}')
+                sdcard.format_device(device=arguments.device, yes=True)
+                sdcard.unmount(device=arguments.device)
+                sdcard.burn_sdcard(tag=tag, device=arguments.device, yes=True)
+                sdcard.mount(device=arguments.device, card_os="ubuntu")
+                if service == 'manager':
+                    # Generate a private public key pair for the manager that will be persistently used
+                    priv_key, pub_key = c.generate_ssh_key(name)
+                    # Write priv_key and pub_key to /boot/id_rsa and /boot/id_rsa.pub
+                    SDCard.writefile(filename=f'{sdcard.boot_volume}/id_rsa', content=priv_key)
+                    SDCard.writefile(filename=f'{sdcard.boot_volume}/id_rsa.pub', content=pub_key)
+                    c.build_user_data(name=name,
+                                      country=arguments.country,
+                                      upgrade=arguments.upgrade,
+                                      with_bridge=enable_bridge).write(
+                        filename=sdcard.boot_volume + '/user-data')
+                    c.build_network_data(name=name, ssid=arguments.ssid, password=arguments.wifipassword)\
+                        .write(filename=sdcard.boot_volume + '/network-config')
+                else:
+                    c.build_user_data(name=name, add_manager_key=manager,
+                                      upgrade=arguments.upgrade).write(
+                        filename=sdcard.boot_volume + '/user-data')
+                    c.build_network_data(name=name).write(filename=sdcard.boot_volume + '/network-config')
+                time.sleep(1)  # Sleep for 1 seconds to give ample time for writing to finish
+                sdcard.unmount(device=arguments.device, card_os="ubuntu")
+
+                Console.info("Remove card")
+
+            Console.ok(f"Burned {len(names)} card(s)")
+            return ""
+
         elif arguments.firmware and arguments.check:
 
             execute("firmware check", burner.firmware(action="check"))
@@ -484,7 +665,7 @@ class BurnCommand(PluginCommand):
         elif arguments.wifi:
 
             password = arguments.passwd
-            ssid = arguments.ssid
+            ssid = arguments.ssid or get_ssid()
             country = arguments.country
 
             if password is None:
@@ -571,6 +752,8 @@ class BurnCommand(PluginCommand):
                 return ""
 
             arguments.TAG = arguments.TAG or ["latest-lite"]
+            if any("ubuntu" in tag for tag in arguments.TAG):
+                sdcard = SDCard(card_os="ubuntu")
 
             execute("format", sdcard.format_device(device=arguments.device, unmount=True))
             execute("unmount", sdcard.unmount(device=arguments.device))
@@ -591,7 +774,8 @@ class BurnCommand(PluginCommand):
 
         elif arguments.unmount:
 
-            execute("unmount", sdcard.unmount(device=arguments.device, card_os=arguments.os))
+            card = SDCard(card_os=arguments.os)
+            execute("unmount", card.unmount(device=arguments.device, card_os=arguments.os))
             return ""
 
         elif arguments.mac:
@@ -623,6 +807,9 @@ class BurnCommand(PluginCommand):
             if arguments.keyboard:
                 execute("set keyboard", burner.keyboard(country=arguments.keyboard))
 
+            if arguments.cmdline:
+                execute("set cmdline", burner.set_cmdline(arguments.cmdline))
+
             return ""
 
         elif arguments.enable and arguments.ssh:
@@ -814,3 +1001,57 @@ class BurnCommand(PluginCommand):
 
         Console.error("see manual page: cms help burn")
         return ""
+
+
+def _build_default_inventory(filename, manager, workers, ips=None, images=None):
+    # cms inventory add red --service=manager --ip=10.1.1.1 --tag=latest-lite
+    # --timezone="America/Indiana/Indianapolis" --locale="us"
+    # cms inventory set red services to "bridge" --listvalue
+    # cms inventory add "red0[1-3]" --service=worker --ip="10.1.1.[2-4]"
+    # --router=10.1.1.1 --tag=latest-lite  --timezone="America/Indiana/Indianapolis" --locale="us"
+    # cms inventory set "red0[1-3]" dns to "8.8.8.8,8.8.4.4" --listvalue
+
+    Console.info("No inventory found or forced rebuild. Buidling inventory "
+                 "with defaults.")
+    Shell.execute("rm", arguments=[
+                  '-f', filename])
+    i = Inventory(filename=filename)
+    timezone = Shell.timezone()
+    locale = Shell.locale()
+    manager_ip = ips[0] if ips else '10.1.1.1'
+    image = images[0] if images else 'latest-lite'
+    element = {}
+    element['host'] = manager
+    element['status'] = 'inactive'
+    element['service'] = 'manager'
+    element['ip'] = manager_ip
+    element['tag'] = image
+    element['timezone'] = timezone
+    element['locale'] = locale
+    element['services'] = ['bridge', 'wifi']
+    element['keyfile'] = '~/.ssh/id_rsa.pub'
+    i.add(**element)
+    i.save()
+
+    last_octet = 2
+    index = 1
+    for worker in workers:
+        ip = ips[index] if ips else f'10.1.1.{last_octet}'
+        image = images[index] if images else 'latest-lite'
+        element = {}
+        element['host'] = worker
+        element['status'] = 'inactive'
+        element['service'] = 'worker'
+        element['ip'] = ip
+        element['tag'] = image
+        element['timezone'] = timezone
+        element['locale'] = locale
+        element['router'] = manager_ip
+        element['dns'] = ['8.8.8.8', '8.8.4.4']
+        element['keyfile'] = '~/.ssh/id_rsa.pub'
+        i.add(**element)
+        i.save()
+        last_octet += 1
+        index += 1
+
+    print(i.list(format="table"))
